/* $Copyright: $
 * Copyright (c) 1995 by Steve Baker (ice@mama.indstate.edu)
 * All Rights reserved
 *
 * This software is provided as is without any express or implied
 * warranties, including, without limitation, the implied warranties
 * of merchantability and fitness for a particular purpose.
 */
#include <utmp.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <paths.h>

static char *version = "$Version: $ sac v1.2 (c) 1995 by Steve Baker $";

/* FSSTND compliance curtesy of Edward S. Marshall */
#ifndef _PATH_WTMP
#define _PATH_WTMP "/var/log/wtmp" /* FSSTND compliant */
#endif

#define min(a,b)	((a) < (b) ? (a) : (b))
#define max(a,b)	((a) > (b) ? (a) : (b))
enum {
  TOTAL	  = 0x01,	/* Total login time */
  DAY     = 0x02,	/* Daily average */
  USER    = 0x03,	/* Per user login time */
  AVERAGE = 0x10,	/* Average login time / login */
  HOUR    = 0x20,	/* Hourly profile */
  CLIP	  = 0x40	/* Multiple logins during same time count only once */
};
enum { FALSE=0, TRUE=1 };
enum { USERLIST, EXCLUDELIST, TTYLIST };

struct day *mkday();
struct usr *adduser(), *finduser();
void *malloc(), *realloc(), *malloc();
time_t getdate(), churn_time();

struct day {
  time_t start, stop;
  int day, month, year;
  time_t time;
  time_t h[24], logins;
  struct day *nxt;
} *days = NULL, *end=NULL;
/*
 * Keep a pointer at the end of the days list, since we'll usually be
 * adding crap to it, not to days before.  What 'o what do we do about
 * time changes aka time warps?
 */

struct usr {
  char user[9];
  time_t time;
  time_t h[24], logins;
  struct usr *nxt;
} *us = NULL;

struct exc {
  char user[9];
  struct exc *nxt;
} *ex = NULL;

struct ttys {
  char line[12];
  struct ttys *nxt;
} *tty = NULL;

struct user {
  char user[9];
  char line[12];
  time_t in;
  struct user *nxt;
} *usr = NULL;

struct tacacs_utmp {
  char ut_line[8];
  char ut_user[8];
  char ut_host[16];
  time_t ut_time;
};

char type = TOTAL, fix = FALSE, exclude = FALSE, fixtty = FALSE, *back = NULL;
int fd;
time_t total = 0, sd = 0, ed = 0, backtime = 0, curtime = 0;
int ndays = 0;
signed int sm = 0, em = 0;

/*
 * sac [-w wtmp] [-dahpcot] [-b hour[:min[:sec]]] [-s start] [-e end]
 *     [[-u] users] [-x [users]] [-T [ttylist]]
 */
main(argc,argv)
int argc;
char **argv;
{
  char *file = _PATH_WTMP, *start = NULL, *end = NULL, old = FALSE, tacacs = FALSE;
  char listtype = USERLIST;
  int i, j, n;
  time_t t, s, e;

  for(n=i=1;i<argc;i=n) {
    n++;
    if (argv[i][0] == '-') {
      for(j=1;argv[i][j];j++) {
	switch (argv[i][j]) {
	  case 'w':
	    file = argv[n++];
	    break;
	  case 'p':
	    type = USER | (type & 0xf0);
	    break;
	  case 'd':
	    type = DAY | (type & 0xf0);
	    break;
	  case 'a':
	    type |= AVERAGE;
	    break;
	  case 'c':
	    type |= CLIP;
	    break;
	  case 'h':
	    type |= HOUR;
	    break;
	  case 'o':
	    old = TRUE;
	    break;
	  case 't':
	    tacacs = TRUE;
	    break;
	  case 'b':
	    back = argv[n++];
	    break;
	  case 's':
	    start = argv[n++];
	    break;
	  case 'e':
	    end = argv[n++];
	    break;
	  case 'x':
	    listtype = EXCLUDELIST;
	    break;
	  case 'T':
	    listtype = TTYLIST;
	    break;
	  case 'u':
	    listtype = USERLIST;
	    break;
	  case '-':
	    if (j == 1) {
	      if (!strcmp("--help",argv[i])) usage(4);
	    }
	  default:
	    usage(1);
	}
      }
    } else {
      switch(listtype) {
	case USERLIST:
	  adduser(argv[i]);
	  fix = TRUE;
	  break;
	case EXCLUDELIST:
	  addexclude(argv[i]);
	  exclude = TRUE;
	  break;
	case TTYLIST:
	  addtty(argv[i]);
	  fixtty = TRUE;
	  break;
      }
    }
  }

  if (start) {
    if (start[0] == '-' || start[0] == '+') {
      for(i=1;start[i];i++) if (!isdigit(start[i])) usage(2);
      sm = atoi(start);
    } else if (isdigit(start[0])) {
      sd = getdate(start);
    } else usage(2);
  }
  if (end) {
    if (end[0] == '-' || end[0] == '+') {
      t = time(0);
      for(i=1;end[i];i++) if (!isdigit(end[i])) usage(2);
      em = atoi(end);
    } else if (isdigit(end[0])) {
      ed = getdate(end);
    } else usage(2);
  }
  if (back) backtime = (curtime=time(0)) - churn_time(back);

  fd = open(file,O_RDONLY);

  if (tacacs) doittacacs();
  else if (old) doitold();
  else doit();

  cleanup();
  report();

  close(fd);
}

usage(n)
int n;
{
  switch (n) {
    case 1:
      fprintf(stderr,"usage: sac [-w wtmp] [-dpahcot] [-s start] [-e end] [-b H[:M[:S]]]\n       [[-u] userlist] [-x [userlist]] [-T [ttylist]]\n");
      break;
    case 2:
      fprintf(stderr,"sac: Invalid date.  Format: +days | -days | mm/dd/yy\n");
      break;
    case 3:
      fprintf(stderr,"sac: Invalid time.  Format: hours[:minutes[:seconds]]\n");
      break;
    case 4:
      fprintf(stderr,"usage: sac [-w wtmp] [-dpahcot] [-s start] [-e end] [-b H[:M[:S]]]\n       [[-u] userlist] [-x [userlist]] [-T [ttylist]]\n");
      fprintf(stderr,"    -w wtmp     Read alternate wtmp file.\n");
      fprintf(stderr,"    -d          Perform daily accounting.\n");
      fprintf(stderr,"    -a          Average usage / login.\n");
      fprintf(stderr,"    -h          Show hourly profile.\n");
      fprintf(stderr,"    -p          Per user accounting.\n");
      fprintf(stderr,"    -c          Perform login clipping.\n");
      fprintf(stderr,"    -o          Read old decrepid wtmp format.\n");
      fprintf(stderr,"    -t          Read tacacs wtmp format.\n");
      fprintf(stderr,"    -s start    Display accounting info from `start'.\n");
      fprintf(stderr,"    -e end      Display accounting info up to `end'.\n");
      fprintf(stderr,"    -b H:M:S    Show accounting info from the last few hours/minutes/seconds.\n");
      fprintf(stderr,"    -x [users]  Does not perform accounting for [users].\n");
      fprintf(stderr,"    -T [ttys]   Perform accounting on only those ttys listed.\n");
      fprintf(stderr,"    -u [users]  Perform accounting only for the users listed.\n");
      fprintf(stderr,"    userlist    Perform accounting only for the users listed.\n");
      break;
  }
  exit(1);
}

/*
 * Turn "mm/dd/yy" into time in seconds.
 */
time_t getdate(s)
char *s;
{
  struct tm tm;
  int y;
  time_t t;

/*
 * <puke>
 * Need a real date parser that can handle different formats and seperators
 */
  if (!isdigit(s[0]) || !isdigit(s[1]) || isdigit(s[2])) usage(2);
  if (!isdigit(s[3]) || !isdigit(s[4]) || isdigit(s[5])) usage(2);
  if (!isdigit(s[6]) || !isdigit(s[7]) || s[8]) usage(2);

  tm.tm_mon = atoi(s) -1;
  tm.tm_mday = atoi(s+3);
  y = atoi(s+6);
  tm.tm_year = (y < 70? 100 + y : y);
  tm.tm_isdst = tm.tm_hour = tm.tm_min = tm.tm_sec = 0;

  t = mktime(&tm);
  if (t == (time_t)(-1)) usage(2);
  return t;
}

time_t churn_time(s)
char *s;
{
  time_t t = 0, mult=3600;
  char nbuf[20], p;

  for(p=0;*s;s++) {
    if (*s == ':') {
      nbuf[p] = 0;
      t += atoi(nbuf) * mult;
      p = 0;
      if (mult > 1) mult /= 60;
      else usage(3);
    } else if (isdigit(*s)) {
      if (p < 15) nbuf[p++] = *s;
      else usage(3);
    } else usage(3);
  }
  nbuf[p] = 0;
  t += atoi(nbuf) * mult;

  return t;
}

doit()
{
  struct utmp u;

  while (read(fd,&u,sizeof(struct utmp)) == sizeof(struct utmp)) {
    if (u.ut_time == 0) continue;	/* corrupted wtmp entry! */
    checkday(u);  /* Do we have this day allocated? */
    /* Q: Why does the following bother me? */
    /* A: It may not handle all possible cases. Wtmp documentation sucks.
     *    Programs are also pretty free to put whatever the hell they want
     *    in wtmp.
     */
    if (u.ut_line[0]) {
      if (u.ut_user[0]) {
	if (!strcmp("~",u.ut_line) && (!strcmp("reboot",u.ut_user) || !strcmp("shutdown",u.ut_user))) do_reboot(u);
	else if (u.ut_type == USER_PROCESS && strncmp("ftp",u.ut_line,3)) login(u);
	else if (u.ut_type == DEAD_PROCESS && strncmp("ftp",u.ut_line,3)) logout(u,0);
	else if (u.ut_type == LOGIN_PROCESS) logout(u,0);
      } else {
	if (!strcmp("|",u.ut_line) || !strcmp("{",u.ut_line)) changetime(u);
	else logout(u,0);
      }
    }
  }
}

/*
 * Do it the old way, where u.ut_type is invalid or doesn't get used.
 * This is for non-standard wtmp's kept by programs such as tacacs.
 * Note: Xterm writes a wtmp entry on exit that looks like a login entry
 *	 if we ignore the ut_type field (which is "dead process" in this
 *	 case).  This will obviously screw things up.
 *
 * This may not be 100%.
 */
doitold()
{
  struct utmp u;

  while (read(fd,&u,sizeof(struct utmp)) == sizeof(struct utmp)) {
    if (u.ut_time == 0) continue;	/* corrupted wtmp entry! */
    checkday(u);
    if (u.ut_line[0]) {
      if (u.ut_user[0]) {
	if (!strcmp("~",u.ut_line) && (!strcmp("reboot",u.ut_user) || !strcmp("shutdown",u.ut_user))) do_reboot(u);
	else if (strcmp("LOGIN",u.ut_user) && strcmp("~",u.ut_line) && strncmp("ftp",u.ut_line,3)) login(u);
	else logout(u,0);
      } else {
	if (!strcmp("|",u.ut_line) || !strcmp("{",u.ut_line)) changetime(u);
	else logout(u,0);
      }
    }
  }
}

/*
 * Do it the tacacs way, with tacacs ancient utmp format, which may work for
 * other acient programs.
 * This too may not be 100%.
 */
doittacacs()
{
  struct tacacs_utmp t;
  struct utmp u;

  while (read(fd,&t,sizeof(struct tacacs_utmp)) == sizeof(struct tacacs_utmp)) {
    if (t.ut_time == 0) continue;	/* corrupted wtmp entry! */

    /* put tacacs_utmp into a normal utmp for the rest of this */
    u.ut_time = t.ut_time;
    strncpy(u.ut_line,t.ut_line,8);
    strncpy(u.ut_user,t.ut_user,8);

    checkday(u);
    if (u.ut_line[0]) {
      if (u.ut_user[0]) {
	if (!strcmp("~",u.ut_line) && (!strcmp("reboot",u.ut_user) || !strcmp("shutdown",u.ut_user))) do_reboot(u);
	else if (strcmp("LOGIN",u.ut_user) && strcmp("~",u.ut_line) && strncmp("ftp",u.ut_line,3)) login(u);
	else logout(u,0);
      } else {
	if (!strcmp("|",u.ut_line) || !strcmp("{",u.ut_line)) changetime(u);
	else logout(u,0);
      }
    }
  }
}

/*
 * Make sure the day that the wtmp entry corresponds to, exists in our
 * days list.  If day is less than the starting day by more than a day, or
 * greater than the last day by more than a day, allocate all the days in
 * between as well.
 */
checkday(u)
struct utmp u;
{
  struct day *d, *p;

  if (days == NULL) end = days = mkday(u.ut_time);
  else {
    if (u.ut_time < days->start) {
      p = d = mkday(u.ut_time);
      while (p->stop+1 < days->start) {
        p->nxt = mkday(p->stop+1);
	p = p->nxt;
      }
      p->nxt = days;
      days = d;
    } else if (u.ut_time > end->stop) {
      p = d = mkday(end->stop+1);
      while (p->stop < u.ut_time) {
        p->nxt = mkday(p->stop+1);
	p = p->nxt;
      }
      end->nxt = d;
      end = p;
    }
  }
}

/*
 * Makes a day entry.  We'll assume a day is exactly 24 hours or 86400
 * seconds long.  I don't know if this is the case or not.  I don't think
 * the time algorithm takes into account leap seconds.
 */
struct day *mkday(t)
time_t t;
{
  struct day *d;
  struct tm *tm;
  int i;

  d = malloc(sizeof(struct day));
  tm = localtime(&t);
  tm->tm_hour = tm->tm_min = tm->tm_sec = 0;

  d->start = mktime(tm);
  d->stop = d->start + 86399;
  d->nxt = NULL;
  d->day = tm->tm_mday;
  d->month = tm->tm_mon;
  d->year = tm->tm_year;
  d->logins = d->time = 0;
  for(i=0;i<24;i++) d->h[i] = 0;
  ndays++;
  return d;
}

login(u)
struct utmp u;
{
  struct user *q;

  /*
   * If we still have a login on this line, it's logged out for sure now!
   * Wtmp could be corrupted.
   */
  logout(u,1);

  q = malloc(sizeof(struct user));
  strncpy(q->line,u.ut_line,12);
  strncpy(q->user,u.ut_user,8);
  q->user[8] = 0;
  q->in = u.ut_time;
  q->nxt = usr;
  usr = q;
}

logout(u,f)
struct utmp u;
char f;
{
  struct user *p, *q, ux;

  for(p=q=usr;p;) {
    if (!strcmp(u.ut_line,p->line)) {
/*      if (f) printf("Forced logout for %s on %s\n",p->user,p->line); */
      release(p,u.ut_time);
      if (p == usr) {
	usr = p->nxt;
	free(p);
	p = q = usr;
      } else {
        q->nxt = p->nxt;
	free(p);
	p = q->nxt;
      }
      continue;
    }
    q = p;
    p = p->nxt;
  }
}

/*
 * logout everyone on reboots or crashes.
 */
do_reboot(u)
struct utmp u;
{
  struct user *p, *q;

  for(p=usr;p;) {
    release(p,u.ut_time);
    q = p;
    p=p->nxt;
    free(q);
  }
  usr = NULL;
}

/*
 * A utmp entry denoted with a line of "|" denotes the "old" time before
 * netdate updated the time. An entry with line of "{" denotes the new time.
 * The difference in time is applied to the login time of every user still
 * logged in.
 */
changetime(u)
struct utmp u;
{
  static time_t old;
  struct user *p;
  signed long dif;

  if (!strcmp("|",u.ut_line)) {
    old = u.ut_time;
    return;
  }
  dif = (signed long)(u.ut_time - old);
  for(p=usr;p;p=p->nxt) p->in -= dif;
}

/*
 * Apply login time for users who haven't logged out yet (or that the wtmp file
 * in particular indicates haven't logged out by the EOF) to the days that
 * are in days list.  Login time is not applied to days not listed in the
 * wtmp file (therefor all the days between the first and last wtmp entries).
 * The reason for this is that if you're inspecting an old wtmp file, the
 * wtmp may not indicate all logouts for the last day.  It is not possible
 * to know when the wtmp file really truly ends however, so we apply the
 * minimum of the current time or the ending time for the last day.
 */
cleanup()
{
  struct user *p;
  time_t t = time(0);

  for(p=usr;p;p=p->nxt)
    release(p,t);
}

/*
 * Release a login entry, applying the login time to the particular day
 * entries.
 * A user is logged in on a particular day when:
 *   in  >= start  &&  in  <= stop ||
 *   out >= start  &&  out <= stop ||
 *   in  <  start  &&  out >  stop
 */
release(u,t)
struct user *u;
time_t t;
{
  struct day *p;
  struct usr *up;
  struct user *q;
  signed long tx;
  int i;

  if ((signed long)((t - u->in) + 1) < 0) return;

  /*
   * Clipping assumes that time moves forward. Only in computers is this
   * not necessarily a safe assumption.
   * Clipping rules:
   *   If (login time < all other login times)
   *      reduce logout time (t) to least login time (-1).
   *   else clip entirely.
   */
  if ((type & CLIP) == CLIP) {
    for(q=usr;q;q=q->nxt) {
      if (q != u && !strcmp(u->user,q->user)) {
	/* throw it out if he's already logged in earlier */
	if (q->in <= u->in) return;
	/* reduce the logout time to the lowest login time. */
	if (q->in < t) t = q->in-1;
      }
    }
  }

  if (backtime) {
    if (u->in > curtime || t < backtime) return;
    if (u->in < backtime && t > backtime) u->in = backtime;
    if (u->in < curtime && t > curtime) t = curtime;
  }

/*
 * If we've fixed the user list, then we only want to apply the login time
 * if that user is in our user list.
 */
  if (fix && (up = finduser(u->user)) == NULL) return;

/*
 * Check the exclude user list.
 */
  if (exclude && isexcluded(u->user)) return;

/*
 * Check the tty list.
 */
  if (fixtty && !istty(u->line)) return;

  /* if we're logging usage / user then apply login time to user entry. */
  if ((type & 0x0f) == USER && (up = finduser(u->user))) {
    up->time += (t - u->in) + 1;
    up->logins++;
    if ((type & HOUR) == HOUR) {
      if (u->in < end->start) {
	for(p=days;p;p=p->nxt) {
	  if (u->in >= p->start && u->in <= p->stop) apply_hours(u->in,t,p->start,up->h);
	  else if (t >= p->start && t <= p->stop) apply_hours(u->in,t,p->start,up->h);
	  else if (u->in < p->start && t > p->stop) for(i=0;i<24;i++) up->h[i] += 3600;
	}
      } else apply_hours(u->in,t,end->start,up->h);
    }
    return;
  }

/*
 * We go through this for both daily and total. The total will be accumulated
 * at the end since we don't know the starting and ending dates until we're
 * done.
 */
  if (u->in < end->start) {
    /* Ugh, it's probably yesterday, but we've got to start all the way at
       the beginning. */
    for(p=days;p;p=p->nxt) {
      if (u->in >= p->start && u->in <= p->stop) {
	p->time += (min(p->stop,t) - u->in) + 1;
	p->logins++;
	if ((type & HOUR) == HOUR) apply_hours(u->in,t,p->start,p->h);
      } else if (t >= p->start && t <= p->stop) {
	p->time += (t - max(p->start,u->in)) + 1;
	p->logins++;
	if ((type & HOUR) == HOUR) apply_hours(u->in,t,p->start,p->h);
      } else if (u->in < p->start && t > p->stop) {
	p->time += 86400;
	p->logins++;
	if ((type & HOUR) == HOUR)
	  for(i=0;i<24;i++) p->h[i] += 3600;
      }
    }
  } else {
    end->time += (min(end->stop,t) - max(end->start,u->in)) + 1;
    end->logins++;
    if ((type & HOUR) == HOUR) apply_hours(u->in,t,end->start,end->h);
  }
}

apply_hours(in,out,start,h)
time_t in, out, start, h[24];
{
  int i;
  time_t b, e;

  b = start;
  e = start + 3599;
  for(i=0;i<24;i++) {
    if (in >= b && in <= e) h[i] += (min(e,out) - in) + 1;
    else if (out >= b && out <= e) h[i] += (out - max(b,in)) + 1;
    else if (in < b && out > e) h[i] += 3600;

    b += 3600;
    e += 3600;
  }
}

struct usr *adduser(s)
char *s;
{
  struct usr *u;

  u = malloc(sizeof(struct usr));
  strncpy(u->user,s,8);
  u->user[8] = 0;
  u->time = 0;
  u->nxt = us;
  us = u;
  return us;
}

struct usr *finduser(s)
char *s;
{
  struct usr *u;

  for(u=us;u;u=u->nxt)
    if (!strcmp(s,u->user)) return u;
  if (!fix) return adduser(s);
  else return NULL;
}

addexclude(s)
char *s;
{
  struct exc *u;

  u = malloc(sizeof(struct exc));
  strncpy(u->user,s,8);
  u->user[8] = 0;
  u->nxt = ex;
  ex = u;
  return;
}

int isexcluded(s)
char *s;
{
  struct exc *u;

  for(u=ex;u;u=u->nxt)
    if (!strcmp(s,u->user)) return TRUE;

  return FALSE;
}

addtty(s)
char *s;
{
  struct ttys *t;

  t = malloc(sizeof(struct ttys));
  strncpy(t->line,s,11);
  t->line[11] = 0;
  t->nxt = tty;
  tty = t;
  return;
}

int istty(s)
char *s;
{
  struct ttys *t;

  for(t=tty;t;t=t->nxt)
    if (!strcmp(s,t->line)) return TRUE;

  return FALSE;
}

report()
{
  static char *month[] = {
    "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"
  };
  struct day *d;
  struct usr *u;
  time_t h[24];
  int i;

  /*
   * Narrow down the days list to the range of days specified by the -s and -e
   * options.  This seems kludged in to me, but I really can't see how else to
   * do it w/o two passes over the wtmp file (which would make this work
   * better, i.e. we would know how many logins for the total period and the
   * -s and -e options would apply to user login time).
   * Perhaps just reading the first and last entries (using seek) of the wtmp
   * file and extrapolating the range from that (assuming that the first and
   * last entries are truely valid).
   */
  if ((type & 0x0f) != USER && (sd || ed || sm || em)) {
    if (sm < 0) sd = end->start + (86400*sm);
    else if (sm > 0) sd = days->start + (86400*sm);
    else if (!sd) sd = days->start;

    if (em < 0) ed = end->start + (86400*em);
    else if (em > 0) ed = days->start + (86400*em);
    else if (!ed) ed = end->start;

    if (sd > ed) exit(0);
    for(d=days;d;d=d->nxt) {
      if (d->start == sd) days = d;
      if (d->start == ed) d->nxt = NULL;
    }
  }

  /* Produce the reports... */
  switch(type & 0x0f) {
    case TOTAL:
      for(ndays=total=0,d=days;d;d=d->nxt) {
	total += d->time;
	ndays++;
      }
      if (backtime) printf("Total: %12.2f over %s hours.\n", (float)total/3600, back);
      else printf("Total: %12.2f over %d days.\n", (float)total/3600, ndays);
      if ((type & AVERAGE) == AVERAGE) printf("Average: %10.2f / day.\n",((float)total/3600) / ndays);
      if ((type & HOUR) == HOUR) {
	for(i=0;i<24;i++) h[i] = 0;
	for(d=days;d;d=d->nxt)
	  for(i=0;i<24;i++) h[i] += d->h[i];
	print_hours(h,total);
      }
      break;
    case DAY:
      for(d=days;d;d=d->nxt) {
	printf("%s %2d  total %10.2f",month[d->month],d->day,(float)d->time/3600);
	if ((type & AVERAGE) == AVERAGE) {
	  if (d->logins)
	    printf("\t%5d logins, %10.2f hours / login.",d->logins,((float)d->time/3600)/d->logins);
	  else
	    printf("\t   no logins");
	}
	putchar('\n');
	if ((type & HOUR) == HOUR) print_hours(d->h,d->time);
      }
      break;
    case USER:
      for(u=us;u;u=u->nxt) {
	printf("\t%-8s %10.2f",u->user,(float)u->time/3600);
	if ((type & AVERAGE) == AVERAGE) {
	  if (u->logins)
	    printf("\t%5d logins, %10.2f hours / login.",u->logins,((float)u->time/3600) / u->logins);
	  else
	    printf("\t   no logins");
	}
	putchar('\n');
	if ((type & HOUR) == HOUR) print_hours(u->h,u->time);
      }
      break;
  }
}

print_hours(h,total)
time_t h[24], total;
{
  static char *bar = "########################################################################";
  int i, bl = strlen(bar);
  float p[24], scale, maxp;

  if (!total) {
    for(i=0;i<24;i++)
      printf("%02d-: \n",i);
  } else {
    for(i=0;i<24;i++) {
      p[i] = (float)h[i] / (float)total;
      if (p[i] > maxp) maxp = p[i];
    }
    scale = (float)bl / maxp;

    for(i=0;i<24;i++) {
      printf("%02d-: %.*s\n",i,(int)(scale*p[i]),bar);
    }
  }
}
